==> lam-6.5.9.tar.gz: lam-6.5.9/share/mpi/bcast.c <==
/*
 * Copyright (c) 2001-2002 The Trustees of Indiana University.  
 *                         All rights reserved.
 * Copyright (c) 1998-2001 University of Notre Dame. 
 *                         All rights reserved.
 * Copyright (c) 1994-1998 The Ohio State University.  
 *                         All rights reserved.
 * 
 * This file is part of the LAM/MPI software package.  For license
 * information, see the LICENSE file in the top level directory of the
 * LAM/MPI source distribution.
 * 
 *	Ohio Trollius
 *	Copyright 1997 The Ohio State University
 *	NJN/RBD
 *
 *	$Id: bcast.c,v 6.7.2.1 2002/10/09 19:49:05 brbarret Exp $
 *
 *	Function:	- broadcast info to all nodes SIMD-style
 *	Accepts:	- buffer
 *			- count
 *			- datatype
 *			- root
 *			- communicator
 *	Returns:	- MPI_SUCCESS or an MPI error code
 */

#include <lam_config.h>

#include <app_mgmt.h>
#include <blktype.h>
#include <mpi.h>
#include <mpisys.h>
#include <mpitrace.h>
#include <rpisys.h>
#if LAM_WANT_IMPI
#include <impi.h>
#endif

/*
 * local functions
 */
static int		bcast_lin();
static int		bcast_log();
static int		bcast_lin_lamd();
static int		bcast_log_lamd();

/*@

MPI_Bcast - Broadcasts a message from the process with rank "root" to
            all other processes of the group. 

Input/output Parameters:
+ buff - starting address of buffer (choice) 
. count - number of entries in buffer (integer) 
. datatype - data type of buffer (handle) 
. root - rank of broadcast root (integer) 
- comm - communicator (handle) 

.N IMPI_YES

Algorithm:  

For 4 or less ranks, a linear algorithm is used, where rank 0
loops over sending the message to each other rank.

If more than 4 ranks are involved, a tree-based algorithm is used to
send the messages from rank 0.

.N fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_INTERCOMM
.N MPI_ERR_COUNT
.N MPI_ERR_TYPE
.N MPI_ERR_BUFFER
.N MPI_ERR_ROOT

.N ACK
@*/
int MPI_Bcast(void *buff, int count, MPI_Datatype datatype, 
	      int root, MPI_Comm comm)
{
    int			size;			/* group size */
    struct _gps		*p;			/* favourite pointer */

    lam_initerr();
    lam_setfunc(BLKMPIBCAST);
/*
 * Check for invalid arguments.
 */
    if ((comm == MPI_COMM_NULL) || LAM_IS_INTER(comm)) {
	return(lam_errfunc(comm, BLKMPIBCAST, lam_mkerr(MPI_ERR_COMM, 0)));
    }

    if (datatype == MPI_DATATYPE_NULL) {
	return(lam_errfunc(comm, BLKMPIBCAST, lam_mkerr(MPI_ERR_TYPE, 0)));
    }

    if (count < 0) {
	return(lam_errfunc(comm, BLKMPIBCAST, lam_mkerr(MPI_ERR_COUNT, 0)));
    }

    MPI_Comm_size(comm, &size);

    if ((root >= size) || (root < 0)) {
	return(lam_errfunc(comm, BLKMPIBCAST, lam_mkerr(MPI_ERR_ROOT, 0)));
    }

    LAM_TRACE(lam_tr_cffstart(BLKMPIBCAST));
/*
 * Set debugging parameters.
 */
    p = &(comm->c_group->g_procs[root]->p_gps);

    lam_setparam(BLKMPIBCAST,
	    root | (p->gps_grank << 16), (p->gps_node << 16) | p->gps_idx);
/*
 * Check for zero count case.
 */
    if (count == 0) {
	LAM_TRACE(lam_tr_cffend(BLKMPIBCAST, root, comm, datatype, count));
	lam_resetfunc(BLKMPIBCAST);
	return(MPI_SUCCESS);
    }
/*
 * Decide which algorithm to use.
 */
    if (size <= 1) {
	lam_resetfunc(BLKMPIBCAST);
	return(MPI_SUCCESS);
    }
#if LAM_WANT_IMPI
    else if (LAM_IS_IMPI(comm)) {
      return IMPI_Bcast(buff, count, datatype, root, comm);
    } 
#endif
    else if (size <= LAM_COLLMAXLIN) {
	return(RPI_SPLIT(bcast_lin_lamd, bcast_lin,
	    		(buff, count, datatype, root, comm)));
    }
    else {
	return(RPI_SPLIT(bcast_log_lamd, bcast_log,
	    		(buff, count, datatype, root, comm)));
    }
}

/*
 *	bcast_lin
 *
 *	Function:	- broadcast using O(N) algorithm
 *	Accepts:	- same arguments as MPI_Bcast()
 *	Returns:	- MPI_SUCCESS or error code
 */
static int
bcast_lin(buff, count, datatype, root, comm)

void			*buff;
int			count;
MPI_Datatype		datatype;
int			root;
MPI_Comm		comm;

{
    int			i;			/* favourite index */
    int			size;			/* group size */
    int			rank;			/* caller rank */
    int			err;			/* error code */
    MPI_Request		*preq;			/* ptr request */
    MPI_Request		reqs[LAM_COLLMAXLIN];	/* requests */

    MPI_Comm_size(comm, &size);
    MPI_Comm_rank(comm, &rank);

    lam_mkcoll(comm);
/*
 * Non-root receive the data.
 */
    if (rank != root) {
	err = MPI_Recv(buff, count, datatype, root,
	    		BLKMPIBCAST, comm, MPI_STATUS_IGNORE);
	lam_mkpt(comm);

	if (err != MPI_SUCCESS) {
	    return(lam_errfunc(comm, BLKMPIBCAST, err));
	}

	LAM_TRACE(lam_tr_cffend(BLKMPIBCAST, root, comm, datatype, count));
	lam_resetfunc(BLKMPIBCAST);
	return(MPI_SUCCESS);
    }
/*
 * Root sends data to all others.
 */
    for (i = 0, preq = reqs; i < size; ++i) {

	if (i == rank) continue;

	err = MPI_Send_init(buff, count, datatype, i, BLKMPIBCAST,
	    			comm, preq++);

	if (err != MPI_SUCCESS) {
	    lam_mkpt(comm);
	    return(lam_errfunc(comm, BLKMPIBCAST, err));
	}
    }
/*
 * Start and wait on all requests.
 */
    err = MPI_Startall(size - 1, reqs);
    if (err != MPI_SUCCESS) {
	lam_mkpt(comm);
	return(lam_errfunc(comm, BLKMPIBCAST, err));
    }

    err = MPI_Waitall(size - 1, reqs, MPI_STATUSES_IGNORE);
    lam_mkpt(comm);

    if (err != MPI_SUCCESS) {
	return(lam_errfunc(comm, BLKMPIBCAST, err));
    }
/*
 * Free the requests.
 */
    for (i = 0, preq = reqs; i < size; ++i) {

	if (i == rank) continue;

	err = MPI_Request_free(preq);
	if (err != MPI_SUCCESS) {
	    return(lam_errfunc(comm, BLKMPIBCAST, err));
	}

	++preq;
    }

    LAM_TRACE(lam_tr_cffend(BLKMPIBCAST, root, comm, datatype, count));
    lam_resetfunc(BLKMPIBCAST);
    return(MPI_SUCCESS);
}

/*
 *	bcast_log
 *
 *	Function:	- broadcast using O(log(N)) algorithm
 *	Accepts:	- same arguments as MPI_Bcast()
 *	Returns:	- MPI_SUCCESS or error code
 */
static int
bcast_log(buff, count, datatype, root, comm)

void			*buff;
int			count;
MPI_Datatype		datatype;
int			root;
MPI_Comm		comm;

{
    int			i;			/* favourite index */
    int			size;			/* group size */
    int			rank;			/* caller rank */
    int			vrank;			/* caller virtual rank */
    int			peer;			/* peer rank */
    int			dim;			/* cube dimension */
    int			hibit;			/* high ON bit position */
    int			mask;			/* rank bit mask */
    int			err;			/* error code */
    int			nreqs;			/* # requests */
    MPI_Request		*preq;			/* ptr request */
    MPI_Request		reqs[LAM_COLLMAXDIM];	/* requests */

    MPI_Comm_rank(comm, &rank);
    MPI_Comm_size(comm, &size);

    lam_mkcoll(comm);

    vrank = (rank + size - root) % size;

    dim = comm->c_cube_dim;
    hibit = lam_hibit(vrank, dim);
    --dim;
/*
 * Receive data from parent in the tree.
 */
    if (vrank > 0) {
	peer = ((vrank & ~(1 << hibit)) + root) % size;

	err = MPI_Recv(buff, count, datatype, peer,
			BLKMPIBCAST, comm, MPI_STATUS_IGNORE);
	if (err != MPI_SUCCESS) {
	    lam_mkpt(comm);
	    return(lam_errfunc(comm, BLKMPIBCAST, err));
	}
    }
/*
 * Send data to the children.
 */
    preq = reqs;
    nreqs = 0;

    for (i = hibit + 1, mask = 1 << i; i <= dim; ++i, mask <<= 1) {

	peer = vrank | mask;
	if (peer < size) {
	    peer = (peer + root) % size;
	    ++nreqs;

	    err = MPI_Send_init(buff, count, datatype, peer, BLKMPIBCAST,
				comm, preq++);
	    if (err != MPI_SUCCESS) {
		lam_mkpt(comm);
		return(lam_errfunc(comm, BLKMPIBCAST, err));
	    }
	}
    }
/*
 * Start and wait on all requests.
 */
    if (nreqs > 0) {
	err = MPI_Startall(nreqs, reqs);
	if (err != MPI_SUCCESS) {
	    lam_mkpt(comm);
	    return(lam_errfunc(comm, BLKMPIBCAST, err));
	}

	err = MPI_Waitall(nreqs, reqs, MPI_STATUSES_IGNORE);
	if (err != MPI_SUCCESS) {
	    lam_mkpt(comm);
	    return(lam_errfunc(comm, BLKMPIBCAST, err));
	}

	for (i = 0, preq = reqs; i < nreqs; ++i, ++preq) {

	    err = MPI_Request_free(preq);
	    if (err != MPI_SUCCESS) {
		lam_mkpt(comm);
		return(lam_errfunc(comm, BLKMPIBCAST, err));
	    }
	}
    }

    lam_mkpt(comm);
    LAM_TRACE(lam_tr_cffend(BLKMPIBCAST, root, comm, datatype, count));
    lam_resetfunc(BLKMPIBCAST);
    return(MPI_SUCCESS);
}

/*
 *	bcast_lin_lamd
 *
 *	Function:	- LAMD broadcast using O(N) algorithm
 *	Accepts:	- same arguments as MPI_Bcast()
 *	Returns:	- MPI_SUCCESS or error code
 */
static int
bcast_lin_lamd(buff, count, datatype, root, comm)

void			*buff;
int			count;
MPI_Datatype		datatype;
int			root;
MPI_Comm		comm;

{
    int			i;			/* favourite index */
    int			size;			/* group size */
    int			rank;			/* caller rank */
    int			longproto;		/* long protocol */
    int			err;			/* error code */
    int			nreqs;			/* # requests */
    int			nacks;			/* # acknowledgements */
    int			done;			/* # done requests */
    int			idx[LAM_COLLMAXDIM];	/* indices */
    MPI_Request		reqs[LAM_COLLMAXDIM];	/* requests */
    MPI_Status		stats[LAM_COLLMAXDIM];	/* status structures */

    MPI_Comm_size(comm, &size);
    MPI_Comm_rank(comm, &rank);

    longproto = (datatype->dt_size * count) > MAXNMSGLEN;
    lam_mkcoll(comm);
/*
 * Non-root receives the data.
 */
    if (rank != root) {
	if (longproto) {
	    err = MPI_Recv(buff, 0, MPI_BYTE, root,
		    	    BLKMPIBCAST, comm, MPI_STATUS_IGNORE);
	    if (err != MPI_SUCCESS) {
		lam_mkpt(comm);
		return(lam_errfunc(comm, BLKMPIBCAST, err));
	    }

	    err = MPI_Send(buff, 0, MPI_BYTE, root, BLKMPIBCAST, comm);
	    if (err != MPI_SUCCESS) {
		lam_mkpt(comm);
		return(lam_errfunc(comm, BLKMPIBCAST, err));
	    }
	}

	err = MPI_Recv(buff, count, datatype, root,
			BLKMPIBCAST, comm, MPI_STATUS_IGNORE);

	lam_mkpt(comm);
	if (err != MPI_SUCCESS) {
	    return(lam_errfunc(comm, BLKMPIBCAST, err));
	}

	LAM_TRACE(lam_tr_cffend(BLKMPIBCAST, root, comm, datatype, count));
	lam_resetfunc(BLKMPIBCAST);
	return(MPI_SUCCESS);
    }
/*
 * Root sends data to all others.
 */
    if (longproto) {
	nreqs = 0;
	for (i = 0; i < size; ++i) {
	    if (i == rank) continue;

	    err = MPI_Send(buff, 0, MPI_BYTE, i, BLKMPIBCAST, comm);
	    if (err != MPI_SUCCESS) {
		lam_mkpt(comm);
		return(lam_errfunc(comm, BLKMPIBCAST, err));
	    }

	    err = MPI_Irecv(buff, 0, MPI_BYTE,
				i, BLKMPIBCAST, comm, &reqs[nreqs++]);
	    if (err != MPI_SUCCESS) {
		lam_mkpt(comm);
		return(lam_errfunc(comm, BLKMPIBCAST, err));
	    }
	}
/*
 * Wait for primer ack messages and send data.
 */
	nacks = 0;
	while (nacks < nreqs) {
	    err = MPI_Waitsome(nreqs, reqs, &done, idx, stats);

	    if (err != MPI_SUCCESS) {
		lam_mkpt(comm);
		return(lam_errfunc(comm, BLKMPIBCAST, err));
	    }

	    if (done == MPI_UNDEFINED) {
		lam_mkpt(comm);
		return(lam_errfunc(comm, BLKMPIBCAST,
				lam_mkerr(MPI_ERR_UNKNOWN, 0)));
	    }

	    nacks += done;

	    for (i = 0; i < done; i++) {

		err = MPI_Send(buff, count, datatype,
				stats[i].MPI_SOURCE, BLKMPIBCAST, comm);
		
		if (err != MPI_SUCCESS) {
		    lam_mkpt(comm);
		    return(lam_errfunc(comm, BLKMPIBCAST, err));
		}
	    }
	}
    }
/*
 * short protocol
 */
    else {
	for (i = 0; i < size; i++) {
	    if (i == root) continue;

	    err = MPI_Send(buff, count, datatype, i, BLKMPIBCAST, comm);
	    if (err != MPI_SUCCESS) {
		lam_mkpt(comm);
		return(lam_errfunc(comm, BLKMPIBCAST, err));
	    }
	}
    }
    
    lam_mkpt(comm);
    LAM_TRACE(lam_tr_cffend(BLKMPIBCAST, root, comm, datatype, count));
    lam_resetfunc(BLKMPIBCAST);
    return(MPI_SUCCESS);
}

/*
 *	bcast_log_lamd
 *
 *	Function:	- LAMD broadcast using O(log(N)) algorithm
 *	Accepts:	- same arguments as MPI_Bcast()
 *	Returns:	- MPI_SUCCESS or error code
 */
static int
bcast_log_lamd(buff, count, datatype, root, comm)

void			*buff;
int			count;
MPI_Datatype		datatype;
int			root;
MPI_Comm		comm;

{
    int			i;			/* favourite index */
    int			size;			/* group size */
    int			rank;			/* caller rank */
    int			vrank;			/* caller virtual rank */
    int			peer;			/* peer rank */
    int			dim;			/* cube dimension */
    int			hibit;			/* high ON bit position */
    int			mask;			/* rank bit mask */
    int			longproto;		/* long protocol */
    int			err;			/* error code */
    int			nreqs;			/* # requests */
    int			nacks;			/* # acknowledgements */
    int			done;			/* # done requests */
    int			idx[LAM_COLLMAXDIM];	/* indices */
    MPI_Request		reqs[LAM_COLLMAXDIM];	/* requests */
    MPI_Status		stats[LAM_COLLMAXDIM];	/* status structures */

    MPI_Comm_rank(comm, &rank);
    MPI_Comm_size(comm, &size);

    lam_mkcoll(comm);
    longproto = (datatype->dt_size * count) > MAXNMSGLEN;

    vrank = (rank + size - root) % size;

    dim = comm->c_cube_dim;
    hibit = lam_hibit(vrank, dim);
    --dim;
/*
 * Receive data from parent in the tree.
 */
    if (vrank > 0) {
	peer = ((vrank & ~(1 << hibit)) + root) % size;

	if (longproto) {
	    err = MPI_Recv(buff, 0, MPI_BYTE, peer,
		    	    BLKMPIBCAST, comm, MPI_STATUS_IGNORE);
	    if (err != MPI_SUCCESS) {
		lam_mkpt(comm);
		return(lam_errfunc(comm, BLKMPIBCAST, err));
	    }

	    err = MPI_Send(buff, 0, MPI_BYTE, peer, BLKMPIBCAST, comm);
	    if (err != MPI_SUCCESS) {
		lam_mkpt(comm);
		return(lam_errfunc(comm, BLKMPIBCAST, err));
	    }
	}

	err = MPI_Recv(buff, count, datatype, peer,
			BLKMPIBCAST, comm, MPI_STATUS_IGNORE);
	if (err != MPI_SUCCESS) {
	    lam_mkpt(comm);
	    return(lam_errfunc(comm, BLKMPIBCAST, err));
	}
    }
/*
 * Send data or primer message to the children.
 */
    nreqs = 0;

    for (i = hibit + 1, mask = 1 << i; i <= dim; ++i, mask <<= 1) {

	peer = vrank | mask;
	if (peer < size) {
	    peer = (peer + root) % size;

	    if (longproto) {
		err = MPI_Send(buff, 0, MPI_BYTE, peer, BLKMPIBCAST, comm);
		if (err != MPI_SUCCESS) {
		    lam_mkpt(comm);
		    return(lam_errfunc(comm, BLKMPIBCAST, err));
		}

		err = MPI_Irecv(buff, 0, MPI_BYTE,
		    		peer, BLKMPIBCAST, comm, &reqs[nreqs++]);
		
	    } else {
		err = MPI_Send(buff, count, datatype, peer, BLKMPIBCAST, comm);
	    }
	    if (err != MPI_SUCCESS) {
		lam_mkpt(comm);
		return(lam_errfunc(comm, BLKMPIBCAST, err));
	    }
	}
    }

    if (longproto) {
/*
 * Wait for primer ack messages and send data.
 */
	nacks = 0;

	while (nacks < nreqs) {

	    err = MPI_Waitsome(nreqs, reqs, &done, idx, stats);

	    if (err != MPI_SUCCESS) {
		lam_mkpt(comm);
		return(lam_errfunc(comm, BLKMPIBCAST, err));
	    }

	    if (done == MPI_UNDEFINED) {
		lam_mkpt(comm);
		return(lam_errfunc(comm, BLKMPIBCAST,
				lam_mkerr(MPI_ERR_UNKNOWN, 0)));
	    }

	    nacks += done;

	    for (i = 0; i < done; i++) {

		err = MPI_Send(buff, count, datatype,
				stats[i].MPI_SOURCE, BLKMPIBCAST, comm);
		
		if (err != MPI_SUCCESS) {
		    lam_mkpt(comm);
		    return(lam_errfunc(comm, BLKMPIBCAST, err));
		}
	    }
	}
    }
    
    lam_mkpt(comm);
    LAM_TRACE(lam_tr_cffend(BLKMPIBCAST, root, comm, datatype, count));
    lam_resetfunc(BLKMPIBCAST);
    return(MPI_SUCCESS);
}

==> lam_7.0.6.orig.tar.gz: lam-7.0.6/share/mpi/bcast.c <==
/*
 * Copyright (c) 2001-2003 The Trustees of Indiana University.  
 *                         All rights reserved.
 * Copyright (c) 1998-2001 University of Notre Dame. 
 *                         All rights reserved.
 * Copyright (c) 1994-1998 The Ohio State University.  
 *                         All rights reserved.
 * 
 * This file is part of the LAM/MPI software package.  For license
 * information, see the LICENSE file in the top level directory of the
 * LAM/MPI source distribution.
 * 
 * $HEADER$
 *
 * $Id: bcast.c,v 6.13 2003/02/04 18:06:51 jsquyres Exp $
 *
 *	Function:	- broadcast info to all nodes SIMD-style
 *	Accepts:	- buffer
 *			- count
 *			- datatype
 *			- root
 *			- communicator
 *	Returns:	- MPI_SUCCESS or an MPI error code
 */

#include <lam_config.h>

#include <app_mgmt.h>
#include <blktype.h>
#include <mpi.h>
#include <mpisys.h>
#include <mpitrace.h>
#include <rpisys.h>
#if LAM_WANT_IMPI
#include <impi.h>
#endif
#include <lam-ssi-rpi.h>
#include <lam-ssi-coll.h>


/*@

MPI_Bcast - Broadcasts a message from the process with rank "root" to
            all other processes of the group. 

Input/output Parameters:
+ buff - starting address of buffer (choice) 
. count - number of entries in buffer (integer) 
. datatype - data type of buffer (handle) 
. root - rank of broadcast root (integer) 
- comm - communicator (handle) 

.N IMPI_YES

Algorithm:  

For 4 or less ranks, a linear algorithm is used, where rank 0
loops over sending the message to each other rank.

If more than 4 ranks are involved, a tree-based algorithm is used to
send the messages from rank 0.

.N fortran

.N Errors
.N MPI_SUCCESS
.N MPI_ERR_COMM
.N MPI_ERR_COLL_NOT_IMPLEMENTED
.N MPI_ERR_COUNT
.N MPI_ERR_TYPE
.N MPI_ERR_BUFFER
.N MPI_ERR_ROOT

.N ACK
@*/
int MPI_Bcast(void *buff, int count, MPI_Datatype datatype, 
	      int root, MPI_Comm comm)
{
    int size;
    int err;
    struct _gps *p;
    lam_ssi_coll_bcast_fn_t func;

    lam_initerr();
    lam_setfunc(BLKMPIBCAST);
/*
 * Check for invalid arguments.
 */
    if (comm == MPI_COMM_NULL) {
      return(lam_errfunc(comm, BLKMPIBCAST, 
			 lam_mkerr(MPI_ERR_COMM, EINVAL)));
    }
    func = LAM_IS_INTER(comm) ?
      comm->c_ssi_coll.lsca_bcast_inter :
      comm->c_ssi_coll.lsca_bcast_intra;

    if (datatype == MPI_DATATYPE_NULL) {
      return(lam_errfunc(comm, BLKMPIBCAST, 
			 lam_mkerr(MPI_ERR_TYPE, EINVAL)));
    }

    if (count < 0) {
      return(lam_errfunc(comm, BLKMPIBCAST, 
			 lam_mkerr(MPI_ERR_COUNT, EINVAL)));
    }

    MPI_Comm_size(comm, &size);
    if ((root >= size) || (root < 0)) {
      return(lam_errfunc(comm, BLKMPIBCAST, 
			 lam_mkerr(MPI_ERR_ROOT, EINVAL)));
    }

    if (func == NULL) {
      return(lam_errfunc(comm, BLKMPIBCAST,
			 lam_mkerr(MPI_ERR_OTHER, ENOT_IMPLEMENTED)));
    }

    LAM_TRACE(lam_tr_cffstart(BLKMPIBCAST));
/*
 * Set debugging parameters.
 */
    p = &(comm->c_group->g_procs[root]->p_gps);
    lam_setparam(BLKMPIBCAST,
	    root | (p->gps_grank << 16), (p->gps_node << 16) | p->gps_idx);
/*
 * Check for zero count case.
 */
    if (count == 0 && comm->c_ssi_coll.lsca_bcast_optimization) {
	LAM_TRACE(lam_tr_cffend(BLKMPIBCAST, root, comm, datatype, count));
	lam_resetfunc(BLKMPIBCAST);
	return(MPI_SUCCESS);
    }

    /* If there's only one node, we're done */

    else if (size <= 1) {
	LAM_TRACE(lam_tr_cffend(BLKMPIBCAST, root, comm, datatype, count));
	lam_resetfunc(BLKMPIBCAST);
	return(MPI_SUCCESS);
    }

    /* Call the coll SSI to actually perform the barrier */
    
    if ((err = func(buff, count, datatype, root, comm)) != MPI_SUCCESS)
      return(lam_errfunc(comm, BLKMPIBCAST, 
			 lam_mkerr(MPI_ERR_COMM, err)));

    /* All done */
    
    LAM_TRACE(lam_tr_cffend(BLKMPIBCAST, root, comm, datatype, count));
    lam_resetfunc(BLKMPIBCAST);
    return MPI_SUCCESS;
}
